/*
CamineetDemo
    This demo is used to test new concepts for Tiamat framework classes. 
    The objectives are:
        -Extend Area class.
        -Test ideas about game scripting.
        -Test event handling ideas.
        -General UI experimentation.
CamineetDemo.java - main class for the demo
    
 Copyright  (C) 2005 Hugues Johnson
 Phantasy Star Copyright  (C) 1988 Sega
 
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.
 
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
the GNU General Public License for more details.
 
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

package com.huguesjohnson.tiamat.demos.Camineet;

import com.huguesjohnson.FileCopy;
import com.huguesjohnson.FileDelete;
import com.huguesjohnson.ImageLoader;
import com.huguesjohnson.PathResolver;
import com.huguesjohnson.fileFilter.MultiExtensionFileFilter;
import com.huguesjohnson.serialization.StorageException;
import com.huguesjohnson.serialization.XmlFormatter;
import com.huguesjohnson.tiamat.Character;
import com.huguesjohnson.tiamat.CharacterFactory;
import com.huguesjohnson.tiamat.Event;
import com.huguesjohnson.tiamat.GameData;
import com.huguesjohnson.tiamat.GameScript;
import com.huguesjohnson.tiamat.Item;
import com.huguesjohnson.tiamat.PlayerCharacter;
import com.huguesjohnson.tiamat.Selection;
import com.huguesjohnson.tiamat.Sequence;
import com.huguesjohnson.tiamat.SequenceFrame;
import com.huguesjohnson.tiamat.StringTable;
import com.huguesjohnson.tiamat.UnsupportedCharacterException;
import com.huguesjohnson.tiamat.audio.AudioPlayer;
import com.huguesjohnson.tiamat.audio.AudioPlayerFactory;
import com.huguesjohnson.tiamat.audio.UnsupportedAudioFormatException;
import com.huguesjohnson.tiamat.graphics.Theme;
import java.awt.Font;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.util.ArrayList;
import java.util.zip.CRC32;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;

/** CamineetDemo - main class for the demo
 * @author Hugues Johnson
 */
public class CamineetDemo implements Runnable{
    /* here for when I'm trying to debug and listen to Loveline or Overclocked Remix at the same time */
    private final static boolean mute=false;
    /* path to the development configuration file */
//    private final transient static String devConfigFilePath="D:\\Projects\\com\\huguesjohnson\\tiamat\\demos\\Camineet\\config.xml";
    /* path to the configuration file */
    private String configFilePath;
    /* the configuration */
    private StringTable config;
    /* the frame */
    private CamineetDemoFrame frame;
    /* the game state */
    private CDGameState gameState;
    /* the main character */
    private PlayerCharacter player;
    /* path to the audio file that's currently playing */
    private String currentAudioFile;
    /* the audio player */
    private AudioPlayer audioPlayer;
    /* path to the game in progress data directory */
    private String gameInProgressDataPath;
    /* path to the base data directory */
    private String baseDataPath;
    /* the current area */
    private CDArea currentArea;
    /* the game data */
    private GameData gameData;
    /* the game script */
    private GameScript script;
    /* reference to the active sequence if one is running */
    private Sequence activeSequence;
    /* the last path used for a saving or loading a file */
    private String lastOpenPath;
    /* the help text */
    private final static String helpText=
        "Press Page Up and Page Down to scroll this text." +
        "\nPress the arrow keys to change the selected action." +
        "\nPress the enter key to execute the selected action.";
    private final static String copyrightText=
        "Camineet Demo (c) 2005-2007 Hugues Johnson" +
        "\nPhantasy Star (c) 1988 Sega";
    
    /** constructor
     * @param configFilePath The path to the configuration file.
     */
    public CamineetDemo(String configFilePath){
        this.configFilePath=new String(configFilePath);
        this.gameState=new CDGameState();
        this.activeSequence=null;
        Thread thread=new Thread(this);
        thread.start();
    }
    
    /** exitHandler - clean up stuff */
    public void exitHandler(){
        this.destroyGameInProgressDataDirectory();
        this.onExit();
    }
    
    /** onExit - event that fires when exit is complete - override this if creating an instance of CamineetDemo in another class */
    public void onExit(){
        System.exit(0);
    }
    
    /** start the thread */
    @SuppressWarnings("serial")
	public void run(){
        //load the configuration
        this.loadConfiguration();
        if(this.config==null){
            JOptionPane.showMessageDialog(null,"Unable to load configuration file: "+this.configFilePath,"Error",JOptionPane.ERROR_MESSAGE);
        } else{
            //set the value for absolute path that will be used throughout this method
            String absolutePath=System.getProperty("user.dir");
            int indexOf=this.configFilePath.lastIndexOf(File.separator);
            if(indexOf>0){
                absolutePath=this.configFilePath.substring(0,indexOf+File.separator.length());
            }
            //set lastOpenPath to absolutePath
            this.lastOpenPath=new String(absolutePath);
            //create the game in progress data directory
            this.baseDataPath=new String(PathResolver.getAbsolutePath(absolutePath,this.config.get("baseDataPath",".\\data\\")));
            this.createGameInProgressDataDirectory();
            //load the game data & script
            String gameDataPath=PathResolver.getAbsolutePath(this.gameInProgressDataPath,this.config.get("dataFile",".\\camineet.data.xml"));
            XmlFormatter formatter=new XmlFormatter();
            try{
                this.gameData=(GameData)formatter.loadFrom(gameDataPath);
                this.gameData.onDeserialize(formatter,this.gameInProgressDataPath);
                this.script=this.gameData.getScript();
            } catch(StorageException x){
                x.printStackTrace();
                //unrecoverable error
                this.onExit();
            }
            //load the main character
            String mainCharacterPath=this.gameInProgressDataPath+"characters\\"+this.config.get("mainCharacter","Evynn")+".character.xml";
            try{
                this.player=(PlayerCharacter)CharacterFactory.load(mainCharacterPath,formatter);
            } catch(UnsupportedCharacterException ucx){
                ucx.printStackTrace();
                //unrecoverable error
                this.onExit();
            }
            //load the Theme
            String themePath=PathResolver.getAbsolutePath(absolutePath,this.config.get("ui.theme",".\\main.theme.xml"));
            Theme theme=new Theme();
            try{
                theme=(Theme)formatter.loadFrom(themePath);
            } catch(StorageException x){
                x.printStackTrace();
            }
            //load the font
            Font font=new Font(null,Font.PLAIN,14);
            try{
                String fontPath=PathResolver.getAbsolutePath(absolutePath,this.config.get("ui.fontpath",".\\fonts\\VeraBold.ttf"));
                FileInputStream in=new FileInputStream(fontPath);
                font=Font.createFont(Font.TRUETYPE_FONT,in);
                font=font.deriveFont(Font.PLAIN,14);
                in.close();
            } catch(Exception x){
                x.printStackTrace();
            }
            //load the window
            this.frame=new CamineetDemoFrame(theme,font){
                public void onWindowClose(){
                    exitHandler();
                }
                public void postEvent(Event event){
                    processGameEvent(event); 
                }
            };
            this.frame.setVisible(true);

//            //hack for testing, comment out the rest of the routine below for this to work
//            this.gameState.addCondition("HAVELETTER");
//            this.gameState.addCondition("HAVEPASS");
//            this.gameState.addCondition("DIALOG1");
//            this.gameState.addCondition("DIALOG2");
//            this.gameState.addCondition("LEARNPLOT");
//            this.gameState.addCondition("DIALOG3");
//            this.gameState.addCondition("DIALOG4");
//            this.loadArea("Camineet-East");            
            //setup the intial menu
            String imagePath=PathResolver.getAbsolutePath(absolutePath,this.config.get("menuImage",".\\images\\sequences\\intro.gif"));
            this.frame.drawArea(ImageLoader.loadBufferedImage(imagePath),null,null);
            Selection[] selections=new Selection[2];
            Event event=new Event("new-game");
            selections[0]=new Selection("New Game",event);
            event=new Event("load");
            selections[1]=new Selection("Load",event);
            this.frame.setSelections(selections);
            this.frame.drawText(CamineetDemo.helpText+"\n\nPlease select New Game or Load.");
        }
    }

    /* load the game over menu */
    private void loadGameOverMenu(){
        Selection[] selections=new Selection[2];
        Event event=new Event("new-game");
        selections[0]=new Selection("New Game",event);
        event=new Event("quit");
        selections[1]=new Selection("Quit",event);
        this.frame.setSelections(selections);
    }
    
    /* load an area */
    private void loadArea(String areaName){
        String areaPath=this.gameInProgressDataPath+"areas\\"+areaName+".area.xml";
        XmlFormatter formatter=new XmlFormatter();
        try{
            CDArea area=(CDArea)formatter.loadFrom(areaPath);
            this.gameState.setCurrentArea(area);
            //start the audio
            this.playAudioFile(area.getBgMusicPathAbsolute());
            //draw stuff
            //TODO this is a really hacky/hokey way to do this
            BufferedImage bgImage;
            if(this.gameState.getTimeOfDay()==CDGameState.TIME_DAY){
                bgImage=area.getBgImage();
            } else{
                String bgImagePath=new String(area.getBgImagePathAbsolute());
                bgImagePath=bgImagePath.replaceAll("day.gif","night.gif");
                bgImage=ImageLoader.loadBufferedImage(bgImagePath);
                if(bgImage==null){
                    bgImage=area.getBgImage();
                }
            }
            BufferedImage overlayImage=area.getOverlayImage();
            BufferedImage characterImage=null;
            Character character=area.getCharacter(0);
            if(character!=null){
                characterImage=character.getFullImage();
            }
            this.frame.drawArea(bgImage,characterImage,overlayImage);
            //load the menu
            this.loadDefaultMenu();
            //if everything was successful, save a reference to the area
            this.currentArea=area;
            this.gameState.setCurrentAreaName(areaName);
            //fire a look event in the new area
            Event e=new Event("look");
            e.getParameters().set("target","Area");
            e.getParameters().set("area",areaName);
            this.runScriptEvent(e);
        } catch(StorageException x){
            x.printStackTrace();
        }
    }
    
    /* process events */
    public void processGameEvent(Event event){
        String actionText=new String(event.getAction().toLowerCase());
        final String nullString=new String("null");
        //check for each possible event, I wish switch worked with strings..
        if(actionText.equals("look")){
            String target=event.getParameters().get("target",nullString);
            if(target.equals(nullString)){
                ArrayList<Selection> selections=new ArrayList<Selection>();
                Event e=new Event("look");
                e.getParameters().set("target","Area");
                e.getParameters().set("area",this.gameState.getCurrentAreaName());
                selections.add(new Selection("Area",e));
                //add a list of exits
                AreaExit[] exits=this.currentArea.getExits();
                if(exits!=null){
                    int length=exits.length;
                    for(int index=0;index<length;index++){
                        e=new Event("look");
                        e.getParameters().set("area",this.gameState.getCurrentAreaName());
                        String exitName=exits[index].getName();
                        String exitDir=exits[index].getDirection();
                        e.getParameters().set("target","exit:"+exitDir);
                        selections.add(new Selection(exitDir,e));
                    }
                }
                //add characters in the area
                Character character=this.currentArea.getCharacter(0);
                if(character!=null){
                    String characterName=character.getProfile().get("Name",nullString);
                    if(!characterName.equals(nullString)){
                        e=new Event("look");
                        e.getParameters().set("area",this.gameState.getCurrentAreaName());
                        e.getParameters().set("target","character:"+characterName);
                        selections.add(new Selection(characterName,e));
                    }
                }
                //add items in the area
                ItemListEntry[] items=this.currentArea.getItemList();
                if(items!=null){
                    int length=items.length;
                    for(int index=0;index<length;index++){
                        String itemName=items[index].getItem().getName();
                        e=new Event("look");
                        e.getParameters().set("area",this.gameState.getCurrentAreaName());
                        e.getParameters().set("target","item:"+itemName);
                        selections.add(new Selection(itemName,e));
                    }
                }
                this.frame.setSelections(selections.toArray(new Selection[0]));
            } else{
                this.runScriptEvent(event);
            }
        } else if(actionText.equals("move")){
            String target=event.getParameters().get("target",nullString);
            if(target.equals(nullString)){
                //add a list of exits
                AreaExit[] exits=this.currentArea.getExits();
                if(exits!=null){
                    int length=exits.length;
                    Selection[] selections=new Selection[length];
                    for(int index=0;index<length;index++){
                        Event e=new Event("move");
                        e.getParameters().set("area",this.gameState.getCurrentAreaName());
                        String exitName=exits[index].getName();
                        String exitDir=exits[index].getDirection();
                        e.getParameters().set("target","area:"+exitName);
                        selections[index]=new Selection(exitDir,e);
                    }
                    this.frame.setSelections(selections);
                }
            } else{
                this.runScriptEvent(event);
            }
        } else if(actionText.equals("talk")){
            String target=event.getParameters().get("target",nullString);
            if(target.equals(nullString)){
                Character[] characters=this.currentArea.getCharacters();
                int length=characters.length;
                if(length<1){
                    Event e=new Event("draw-text");
                    e.getParameters().set("text","There's no one here to talk to.");
                    this.processGameEvent(e);
                } else{
                    Selection[] selections=new Selection[length];
                    for(int index=0;index<length;index++){
                        Event e=new Event("talk");
                        e.getParameters().set("area",this.gameState.getCurrentAreaName());
                        String name=characters[index].getProfile().get("Name","Some dude");
                        e.getParameters().set("target","character:"+name);
                        selections[index]=new Selection(name,e);
                    }
                    this.frame.setSelections(selections);
                }
            } else{
                this.runScriptEvent(event);
                this.loadDefaultMenu();
            }
        } else if(actionText.equals("take")){
            String target=event.getParameters().get("target",nullString);
            if(target.equals(nullString)){
                ItemListEntry[] items=this.currentArea.getItemList();
                int length=items.length;
                if(length<1){
                    Event e=new Event("draw-text");
                    e.getParameters().set("text","There are no items here you can take.");
                    this.processGameEvent(e);
                } else{
                    ArrayList<Selection> selections=new ArrayList<Selection>();
                    for(int index=0;index<length;index++){
                        if(items[index].getCanTake()){
                            Event e=new Event("take");
                            String name=items[index].getItem().getName();
                            e.getParameters().set("area",this.gameState.getCurrentAreaName());
                            e.getParameters().set("target",name);
                            selections.add(new Selection(name,e));
                        }
                    }
                    if(selections.size()>0){
                        this.frame.setSelections(selections.toArray(new Selection[0]));
                    } else{
                        Event e=new Event("draw-text");
                        e.getParameters().set("text","There are no items here you can take.");
                        this.processGameEvent(e);
                    }
                }
            } else{
                this.runScriptEvent(event);
                this.loadDefaultMenu();
            }
        } else if(actionText.equals("use")){
            String target=event.getParameters().get("item",nullString);
            if(!target.equals(nullString)){
                this.runScriptEvent(event);
            }
            this.loadDefaultMenu();
        } else if(actionText.equals("save")){
            this.saveGame();
        } else if(actionText.equals("load")){
            this.loadGame();
        } else if(actionText.equals("help")){
            //create a StringBuffer containing the help text
            StringBuffer text=new StringBuffer(CamineetDemo.helpText);
            //add link to help page
            text.append("\n\nFor additional help with this game please visit \nhttp://www.huguesjohnson.com/tiamat.html");
            //add a hint -- should put this logic into the script
            text.append("\n\nHint: ");
            String[] c=this.gameState.getConditions();
            int length=c.length;
            ArrayList<String> conditions=new ArrayList<String>(length);
            for(int index=0;index<length;index++){
                conditions.add(c[index].toUpperCase());
            }
            if(conditions.contains("GAMEOVER")){
                text.append("It looks like you're missing some items, better check your inventory.");
            } else if(conditions.contains("NIGHT")){
                text.append("It's time to foil this plot, arm yourself.");
            } else if(conditions.contains("DIALOG4")){
                text.append("You'd better find a place to rest until nightfall.");
            } else if(conditions.contains("DIALOG3")){
                text.append("There's a lab in the tunnel between the Spaceport and Gothic.");
            } else if(conditions.contains("LEARNPLOT")){
                text.append("There's someone in Gothic who can help.");
            } else if(conditions.contains("DIALOG2")){
                text.append("Someone is waiting to talk to you in place where the guards won't see.");
            } else if(conditions.contains("DIALOG1")){
                text.append("There are others who can help you, try to find them.");
            } else if(conditions.contains("HAVEPASS")){
                text.append("Find Nekise's friend in Parolit.");
            } else{
                text.append("Explore everywhere you can.");
            }
            //add the copyright text
            text.append("\n\n");
            text.append(CamineetDemo.copyrightText);
            //post an event to draw the text
            Event e=new Event("draw-text");
            e.getParameters().set("text",text.toString());
            this.processGameEvent(e);
        } else if(actionText.equals("inventory")){
            Item[] items=this.player.getItems().getAll();
            int length;
            if(items!=null){ 
                length=items.length; 
            } else{
                length=0;
            }
            if(length<1){
                Event e=new Event("draw-text");
                e.getParameters().set("text","You are not carrying any items.");
                this.processGameEvent(e);
            } else{
                ArrayList<Selection> selections=new ArrayList<Selection>(length);
                for(int index=0;index<length;index++){
                    Event e=new Event("use");
                    e.getParameters().set("area",this.gameState.getCurrentAreaName());
                    String itemName=items[index].getName();
                    e.getParameters().set("item",itemName);
                    selections.add(new Selection(itemName,e));
                }
                this.frame.setSelections(selections.toArray(new Selection[0]));
            }
        } else if(actionText.equals("move-player")){
            String location=event.getParameters().get("location",nullString);
            if(!location.equals(nullString)){
                this.loadArea(location);
            }
        } else if(actionText.equals("draw-text")){
            String text=event.getParameters().get("text",nullString);
            if(!text.equals(nullString)){
                this.frame.drawText(text);
                this.loadDefaultMenu();
            }
        } else if(actionText.equals("add-item")){
            String itemName=event.getParameters().get("item",nullString);
            if(!itemName.equals(nullString)){
                Item item=this.gameData.getItem(itemName);
                if(item!=null){
                    this.player.getItems().setItem(itemName,item);
                }
            }
        } else if(actionText.equals("move-item")){
            String itemName=event.getParameters().get("item",nullString);
            if(!itemName.equals(nullString)){
                Item item=this.gameData.getItem(itemName);
                if(item!=null){
                    //check if there is an old location to remove the item from
                    String oldLocation=event.getParameters().get("old-location",nullString);
                    if(!oldLocation.equals(nullString)){
                        XmlFormatter formatter=new XmlFormatter();
                        String areaPath=this.gameInProgressDataPath+"areas\\"+oldLocation+".area.xml";
                        try{
                            CDArea area=(CDArea)formatter.loadFrom(areaPath);
                            if(area!=null){
                                ArrayList<ItemListEntry> newItemList=new ArrayList<ItemListEntry>();
                                ItemListEntry[] itemList=area.getItemList();
                                int length=itemList.length;
                                for(int index=0;index<length;index++){
                                    if(!itemList[index].getItem().getName().equals(item.getName())){
                                        newItemList.add(itemList[index]);
                                    }
                                }
                                area.setItemList(newItemList.toArray(new ItemListEntry[0]));
                                formatter.saveTo(area,areaPath);
                                if(oldLocation.equals(this.gameState.getCurrentAreaName())){
                                    this.loadArea(this.gameState.getCurrentAreaName());
                                }
                            }
                        } catch(StorageException x){
                            x.printStackTrace();
                        }
                    }
                    //check if there is a new location to add the item to
                    String newLocation=event.getParameters().get("new-location",nullString);
                    if(!newLocation.equals(nullString)){
                        XmlFormatter formatter=new XmlFormatter();
                        String areaPath=this.gameInProgressDataPath+"areas\\"+newLocation+".area.xml";
                        try{
                            CDArea area=(CDArea)formatter.loadFrom(areaPath);
                            if(area!=null){
                                ItemListEntry[] itemList=area.getItemList();
                                int length=itemList.length;
                                ItemListEntry[] newItemList=new ItemListEntry[length+1];
                                if(length>0){
                                    System.arraycopy(itemList,0,newItemList,0,length);
                                }
                                int x=Integer.valueOf(event.getParameters().get("x","0")).intValue();
                                int y=Integer.valueOf(event.getParameters().get("y","0")).intValue();
                                boolean canTake=Boolean.valueOf(event.getParameters().get("canTake","true")).booleanValue();
                                ItemListEntry entry=new ItemListEntry(x,y,canTake,item);
                                newItemList[length-1]=entry;
                                area.setItemList(newItemList);
                                formatter.saveTo(area,areaPath);
                                if(newLocation.equals(this.gameState.getCurrentAreaName())){
                                    this.loadArea(this.gameState.getCurrentAreaName());
                                }
                            }
                        } catch(StorageException x){
                            x.printStackTrace();
                        }
                    }
                }
            }
        } else if(actionText.equals("move-character")){
            String characterName=event.getParameters().get("character",nullString);
            if(!characterName.equals(nullString)){
                Character character=this.gameData.getCharacter(characterName);
                if(character!=null){
                    //check if there is an old location to remove the character from
                    String oldLocation=event.getParameters().get("old-location",nullString);
                    if(!oldLocation.equals(nullString)){
                        XmlFormatter formatter=new XmlFormatter();
                        String areaPath=this.gameInProgressDataPath+"areas\\"+oldLocation+".area.xml";
                        try{
                            CDArea area=(CDArea)formatter.loadFrom(areaPath);
                            if(area!=null){
                                ArrayList<String> newPaths=new ArrayList<String>();
                                String path="characters\\"+characterName+".character.xml";
                                String[] paths=area.getCharacterPaths();
                                int length=paths.length;
                                for(int index=0;index<length;index++){
                                    if(!paths[index].endsWith(path)){
                                        newPaths.add(paths[index]);
                                    }
                                }
                                area.setCharacters(newPaths.toArray(new String[0]),formatter);
                                formatter.saveTo(area,areaPath);
                                if(oldLocation.equals(this.gameState.getCurrentAreaName())){
                                    this.loadArea(this.gameState.getCurrentAreaName());
                                }
                            }
                        } catch(StorageException x){
                            x.printStackTrace();
                        }
                    }
                    //check if there is a new location to add the character to
                    String newLocation=event.getParameters().get("new-location",nullString);
                    if(!newLocation.equals(nullString)){
                        XmlFormatter formatter=new XmlFormatter();
                        String areaPath=this.gameInProgressDataPath+"areas\\"+newLocation+".area.xml";
                        try{
                            CDArea area=(CDArea)formatter.loadFrom(areaPath);
                            if(area!=null){
                                //String path=this.gameInProgressDataPath+"..\\characters\\"+characterName+".character.xml";
                                String path="..\\characters\\"+characterName+".character.xml";
                                String[] paths=area.getCharacterPaths();
                                int length=paths.length;
                                String[] newPaths=new String[length+1];
                                if(length>0){
                                    System.arraycopy(paths,0,newPaths,0,length);
                                    newPaths[length-1]=path;
                                } else{
                                    newPaths[0]=path;
                                }
                                area.setCharacters(newPaths,formatter);
                                formatter.saveTo(area,areaPath);
                                if(newLocation.equals(this.gameState.getCurrentAreaName())){
                                    this.loadArea(this.gameState.getCurrentAreaName());
                                }
                            }
                        } catch(StorageException x){
                            x.printStackTrace();
                        }
                    }
                }
            }
        } else if(actionText.equals("drop-item")){
            String itemName=event.getParameters().get("item",nullString);
            if(!itemName.equals(nullString)){
                this.player.getItems().removeItem(itemName);
            }
        } else if(actionText.equals("add-condition")){
            String condition=event.getParameters().get("condition",nullString);
            if(!condition.equals(nullString)){
                this.gameState.addCondition(condition);
            }
        } else if(actionText.equals("new-game")){
            Event e=new Event("run-sequence");
            e.getParameters().set("sequence","intro");
            this.processGameEvent(e);
        } else if(actionText.equals("next-sequence-frame")){
            try{
                int frame=Integer.parseInt(event.getParameters().get("frame",nullString));
                if(frame<this.activeSequence.getFrameCount()){
                    //set the menu
                    Selection[] selections=new Selection[1];
                    StringTable parameters=new StringTable();
                    parameters.set("frame",String.valueOf(frame+1)); 
                    Event menuEvent=new Event("next-sequence-frame",parameters);
                    selections[0]=new Selection("Next",menuEvent);
                    this.frame.setSelections(selections);
                    //display the sequence
                    SequenceFrame currentFrame=this.activeSequence.getFrame(frame);
                    this.playAudioFile(currentFrame.getBgMusicPathAbsolute());
                    this.frame.drawArea(currentFrame.getBgImage(),null,null);
                    this.frame.drawText(currentFrame.getText());
                } else{ //run the exit events
                    Event[] exitEvents=this.activeSequence.getExitEvents();
                    int length=exitEvents.length;
                    for(int index=0;index<length;index++){
                        this.processGameEvent(exitEvents[index]);
                    }
                    this.activeSequence=null;
                    //HORRIBLE, HORRIBLE FIX TO A BUG
                    if(exitEvents[length-1].getAction().equals("game-over")){
                        this.loadGameOverMenu();
                    } else{
                        this.loadDefaultMenu();
                    }
                }
            } catch(Exception x){
                x.printStackTrace();
            }
        } else if(actionText.equals("set-time")){
            String newTime=event.getParameters().get("time",nullString);
            if(!newTime.equals(nullString)){
                try{
                    int time=Integer.parseInt(newTime);
                    this.gameState.setTimeOfDay(time);
                } catch(Exception x){
                    x.printStackTrace();
                }
            }
        } else if(actionText.equals("run-sequence")){
            String sequenceName=event.getParameters().get("sequence",nullString);
            XmlFormatter formatter=new XmlFormatter();
            String sequencePath=this.gameInProgressDataPath+"sequences\\"+sequenceName+".sequence.xml";
            try{
                //load the sequence
                this.activeSequence=(Sequence)formatter.loadFrom(sequencePath);
                //set the menu
                Selection[] selections=new Selection[1];
                StringTable parameters=new StringTable();
                parameters.set("frame","1"); 
                Event menuEvent=new Event("next-sequence-frame",parameters);
                selections[0]=new Selection("Next",menuEvent);
                this.frame.setSelections(selections);
                //display the sequence
                SequenceFrame currentFrame=this.activeSequence.getFrame(0);
                this.playAudioFile(currentFrame.getBgMusicPathAbsolute());
                this.frame.drawArea(currentFrame.getBgImage(),null,null);
                this.frame.drawText(currentFrame.getText());
            } catch(StorageException x){
                x.printStackTrace();
            }
        } else if(actionText.equals("quit")){
            this.onExit();
        } else if(actionText.equals("game-over")){
            this.loadGameOverMenu();
        }
    }

    /* run an event against the game script */
    private void runScriptEvent(Event event){
        Event[] returnEvents=this.script.processEvent(event,this.gameState.getConditions());
        if((returnEvents==null)||(returnEvents.length<1)){
            //check for default messages
            String actionText=new String(event.getAction().toLowerCase());
            final String nullString=new String("null");
            if(actionText.equals("use")){
                String itemName=event.getParameters().get("item",nullString);
                Item item=this.gameData.getItem(itemName);
                this.frame.drawText(item.getDescription());
            } else{
                StringBuffer text=new StringBuffer("Unsupported event");
                text.append("\naction=");
                text.append(event.getAction());
                String params[][]=event.getParameters().getAll();
                if(params==null){
                    text.append("\nnull parameters");
                } else{
                    //[0][n]=key for item n, [1][n]=value for item n
                    int length=params[0].length;
                    for(int index=0;index<length;index++){
                        text.append("\n");
                        text.append(params[0][index]);
                        text.append("=");
                        text.append(params[1][index]);
                    }
                }
                this.frame.drawText(text.toString());
            }
        } else {
            int length=returnEvents.length;
            for(int index=0;index<length;index++){
                this.processGameEvent(returnEvents[index]);
            }
        } 
    }
    
    /* load the default menu */
    private void loadDefaultMenu(){
        //this check is a workaround to a bug, need to figure out a better way to do this
        if(this.activeSequence==null){
            Selection[] selections=new Selection[8];
            Event event=new Event("look");
            selections[0]=new Selection("Look",event);
            event=new Event("move");
            selections[1]=new Selection("Move",event);
            event=new Event("talk");
            selections[2]=new Selection("Talk",event);
            event=new Event("take");
            selections[3]=new Selection("Take",event);
            event=new Event("inventory");
            selections[4]=new Selection("Inventory",event);
            event=new Event("save");
            selections[5]=new Selection("Save",event);
            event=new Event("load");
            selections[6]=new Selection("Load",event);
            event=new Event("help");
            selections[7]=new Selection("Help",event);
            this.frame.setSelections(selections);
        }
    }
    
    /* create game in progress data directory */
    private void createGameInProgressDataDirectory(){
        if(this.baseDataPath.endsWith(File.separator)){
            this.gameInProgressDataPath=new String(this.baseDataPath.substring(0,this.baseDataPath.length()-File.separator.length()));
        } else{
            this.gameInProgressDataPath=new String(this.baseDataPath);
        }
        String timeStamp=String.valueOf(System.currentTimeMillis());
        this.gameInProgressDataPath=this.gameInProgressDataPath.concat(timeStamp).concat(File.separator);
        FileCopy.copyDirectory(this.baseDataPath,this.gameInProgressDataPath,true);
    }
    
    /* delete the game in progress data directory */
    private void destroyGameInProgressDataDirectory(){
        FileDelete.deleteDirectory(this.gameInProgressDataPath);
    }

    /* play the specified audio file unless it's already playing */
    private void playAudioFile(String audioFile){
        if(!CamineetDemo.mute){
            if(!audioFile.equals(this.currentAudioFile)){
                try{
                    if(this.audioPlayer!=null){
                        this.audioPlayer.stop();
                        this.audioPlayer=null;
                    }
                    this.audioPlayer=AudioPlayerFactory.getAudioPlayerFor(audioFile);
                    this.audioPlayer.play(audioFile,AudioPlayer.INFINITE_LOOP);
                    this.currentAudioFile=new String(audioFile);
                } catch(UnsupportedAudioFormatException uafx){
                    uafx.printStackTrace();
                }
            }
        }
    }
    
    /* load the configuration file */
    private void loadConfiguration(){
        XmlFormatter formatter=new XmlFormatter();
        try{
            this.config=(StringTable)formatter.loadFrom(this.configFilePath);
        } catch(StorageException x){
            x.printStackTrace();
        }
    }
    
    /** save the game */
    private void saveGame(){
        JFileChooser fileChooser=new JFileChooser();
        //do not allow multiple selections
        fileChooser.setMultiSelectionEnabled(false);
        //only accept specific filters
        fileChooser.setAcceptAllFileFilterUsed(false);
        //set the current directory
        fileChooser.setCurrentDirectory(new File(this.lastOpenPath));
        //add file filters
        String[] extensionList=new String[1];
        extensionList[0]=new String("save");
        fileChooser.setFileFilter(new MultiExtensionFileFilter(extensionList));
        //show the dialog
        int returnValue=fileChooser.showSaveDialog(this.frame);
        //was 'OK' selected?
        if(returnValue==JFileChooser.APPROVE_OPTION){
            String saveFile=fileChooser.getSelectedFile().getPath();
            if(!saveFile.endsWith(".save")){
                saveFile=saveFile+".save";
            }
            XmlFormatter formatter=new XmlFormatter();
            //save the current game state
            try{
                formatter.saveTo(this.gameState,this.gameInProgressDataPath+"state.xml");
                //save the last open path
                this.lastOpenPath=new String(fileChooser.getCurrentDirectory().getAbsolutePath());
            } catch(StorageException x){
                x.printStackTrace();
            }
            //save the main character
            String mainCharacterPath=this.gameInProgressDataPath+"characters\\"+this.config.get("mainCharacter","Evynn")+".character.xml";
            try{
                formatter.saveTo(this.player,mainCharacterPath);
            } catch(StorageException x){
                x.printStackTrace();
            }            
            //TODO ISSUE.. WITH THIS WE CAN ONLY SAVE ONE LEVEL DEEP
            //SAVING MAY TAKE A WHILE.. CONSIDER ADDING PROGRESS
            //list all the files to save
            ArrayList<String> fileList=new ArrayList<String>();
            String[] list=(new File(this.gameInProgressDataPath)).list();
            int length=list.length;
            for(int index=0;index<length;index++){
                String fullPath=this.gameInProgressDataPath+list[index];
                File f=new File(fullPath);
                if(f.isDirectory()){
                    String[] list1=f.list();
                    int length1=list1.length;
                    for(int index1=0;index1<length1;index1++){
                        fileList.add(fullPath+File.separator+list1[index1]);
                    }
                } else{
                    fileList.add(fullPath);
                }
            }
            //now write the zip
            try{
                ZipOutputStream zout=new ZipOutputStream(new FileOutputStream(new File(saveFile)));
                zout.setMethod(ZipOutputStream.DEFLATED);
                int count=fileList.size();
                for(int fileNameIndex=0;fileNameIndex<count;fileNameIndex++){
                    String fileName=fileList.get(fileNameIndex);
                    File file=new File(fileName);
                    byte buffer[]=new byte[1024];
                    int blength;
                    /* generate CRC */
                    CRC32 crc=new CRC32();
                    FileInputStream fin=new FileInputStream(file);
                    while((blength=fin.read(buffer))>-1){
                         crc.update(buffer,0,blength);
                    }
                    fin.close();
                    /* create zip entry */
                    String relativePath=PathResolver.getRelativePath(this.gameInProgressDataPath,fileName);
                    if(relativePath.startsWith("."+File.separator)){
                        relativePath=relativePath.substring(File.separator.length()+1);
                    }
                    ZipEntry entry=new ZipEntry(relativePath);
                    entry.setSize(file.length());
                    entry.setTime(file.lastModified());
                    entry.setCrc(crc.getValue());
                    zout.putNextEntry(entry);
                    fin=new FileInputStream(file);
                    /* write entry to zip file */
                    while((length=fin.read(buffer))>-1){
                        zout.write(buffer,0,length);
                    }
                    fin.close();
                    zout.closeEntry();
               }
               zout.close();
          } catch(Exception x){
               x.printStackTrace();
          }            
        }
    }
    
    /** load a saved game */
    private void loadGame(){
        JFileChooser fileChooser=new JFileChooser();
        //do not allow multiple selections
        fileChooser.setMultiSelectionEnabled(false);
        //only accept specific filters
        fileChooser.setAcceptAllFileFilterUsed(false);
        //set the current directory
        fileChooser.setCurrentDirectory(new File(this.lastOpenPath));
        //add file filters
        String[] extensionList=new String[1];
        //extensionList[0]=new String("save.zip");
        extensionList[0]=new String("save");
        fileChooser.setFileFilter(new MultiExtensionFileFilter(extensionList));
        //show the dialog
        int returnValue=fileChooser.showOpenDialog(this.frame);
        //was 'OK' selected?
        if(returnValue==JFileChooser.APPROVE_OPTION){
            String saveFile=fileChooser.getSelectedFile().getPath();
            //NEED SOME ERROR HANDLING IF A BAD SAVE FILE IS OPENED
            //delete the current game state
            this.destroyGameInProgressDataDirectory();
            //create a new directory
            if(this.baseDataPath.endsWith(File.separator)){
                this.gameInProgressDataPath=new String(this.baseDataPath.substring(0,this.baseDataPath.length()-File.separator.length()));
            } else{
                this.gameInProgressDataPath=new String(this.baseDataPath);
            }
            String timeStamp=String.valueOf(System.currentTimeMillis());
            this.gameInProgressDataPath=this.gameInProgressDataPath.concat(timeStamp).concat(File.separator);
            (new File(this.gameInProgressDataPath)).mkdir();
            //unzip the save file
            try{
                ZipInputStream zin=new ZipInputStream(new FileInputStream(new File(saveFile)));
                ZipEntry entry=zin.getNextEntry();
                while(entry!=null){
                    byte buffer[]=new byte[1024];
                    int length;
                    /* write file */
                    String outPath=new String(PathResolver.getAbsolutePath(this.gameInProgressDataPath,entry.getName()));
                    int lastIndexOf=outPath.lastIndexOf(File.separator);
                    if(lastIndexOf>0){
                        String subDirPath=new String(outPath.substring(0,lastIndexOf));
                        File subDir=new File(subDirPath);
                        if(!subDir.exists()){
                            subDir.mkdir();
                        }
                    }
                    FileOutputStream fout=new FileOutputStream(new File(outPath));
                    while((length=zin.read(buffer))>-1){
                        fout.write(buffer,0,length);
                    }
                    zin.closeEntry();
                    fout.close();
                    entry=zin.getNextEntry();
                }
                zin.close();
            } catch(Exception x){
                x.printStackTrace();
            }
            XmlFormatter formatter=new XmlFormatter();
            //reload the game state
            try{
                this.gameState=(CDGameState)formatter.loadFrom(this.gameInProgressDataPath+"state.xml");
                this.loadArea(this.gameState.getCurrentAreaName());
                //save the last open path
                this.lastOpenPath=new String(fileChooser.getCurrentDirectory().getAbsolutePath());
            } catch(StorageException x){
                x.printStackTrace();
            }
            //reload the main character
            String mainCharacterPath=this.gameInProgressDataPath+"characters\\"+this.config.get("mainCharacter","Evynn")+".character.xml";
            try{
                this.player=(PlayerCharacter)CharacterFactory.load(mainCharacterPath,formatter);
            } catch(UnsupportedCharacterException ucx){
                ucx.printStackTrace();
            }            
        } 
    }
    
    /** entry point for program */
    public static void main(String[] arguments){
        //new CamineetDemo(devConfigFilePath);
    	String dir=System.getProperty("user.dir");
    	if(!dir.endsWith(File.separator)){ dir+=File.separator; }
     	new CamineetDemo(dir+"config.xml");
     }
}